class PixiError(Exception):
    pass


class GoAuthenticate(PixiError):
    pass


class InvalidConfig(PixiError):
    pass


class InvalidURL(PixiError):
    pass


class DownloadFailed(PixiError):
    pass


class DuplicateImage(PixiError):
    pass



from pathlib import Path

import click
from appdirs import user_config_dir, user_data_dir

from pixi.errors import PixiError

CONFIG_DIR = Path(user_config_dir('pixi', 'dazzler'))
DATA_DIR = Path(user_data_dir('pixi', 'dazzler'))


@click.group()
def commandgroup():
    pass


def make_app_directories():
    for name, dir_ in [
        ('configuration', CONFIG_DIR),
        ('data', DATA_DIR),
    ]:
        try:
            dir_.mkdir(mode=0o700, parents=True, exist_ok=True)
        except PermissionError:  # pragma: no cover
            raise PixiError(f'Could not create {name} directory.')



from math import ceil

from pixivapi import Client as BaseClient
from pixivapi import LoginError
from tqdm import tqdm

from pixi.config import Config
from pixi.errors import GoAuthenticate
from pixi.util import rename_duplicate_file


class Client:
    """
    A "proxy singleton" that returns the same PixivClient instance when
    instantiated.
    """

    __pixiv_client = None

    def __new__(cls, authenticate=True):
        if cls.__pixiv_client is None:
            cls.__pixiv_client = _PixivClient(authenticate)
        return cls.__pixiv_client


class _PixivClient(BaseClient):
    """
    A Pixiv client that downloads with progress bars!
    """

    def __init__(self, authenticate=True):
        super().__init__()

        if authenticate:
            config = Config()
            if not config['pixi']['refresh_token']:
                raise GoAuthenticate

            try:
                self.authenticate(config['pixi']['refresh_token'])
            except LoginError:
                raise GoAuthenticate

    def download(self, url, destination, referer='https://pixiv.net'):
        response = self.session.get(
            url, headers={'Referer': referer}, stream=True
        )

        destination = rename_duplicate_file(destination)
        try:
            with destination.open('wb') as f:
                for chunk in tqdm(
                    iterable=response.iter_content(chunk_size=1024),
                    total=ceil(int(response.headers['Content-Length']) / 1024),
                    unit='KB',
                    unit_scale=True,
                ):
                    f.write(chunk)
        except:  # noqa E722
            destination.unlink()
            raise



import sqlite3
import sys
from collections import namedtuple
from contextlib import contextmanager
from pathlib import Path

import click

from pixi import DATA_DIR

Migration = namedtuple('Migration', 'path, version')

DATABASE_PATH = DATA_DIR / 'db.sqlite3'
MIGRATIONS_DIR = Path(__file__).parent / 'migrations'


@contextmanager
def database():
    with sqlite3.connect(str(DATABASE_PATH)) as conn:
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        yield conn, cursor
        cursor.close()


def create_database_if_nonexistent():
    with database() as (conn, cursor):
        cursor.execute(
            """
            SELECT name FROM sqlite_master
            WHERE type='table' AND name='versions'
            """
        )
        if not cursor.fetchone():
            cursor.execute(
                """
                CREATE TABLE versions (
                    version INTEGER,
                    PRIMARY KEY (version)
                )
                """
            )
            conn.commit()


def confirm_database_is_updated():
    if calculate_migrations_needed():
        if not len(sys.argv) == 2 or sys.argv[1] != 'migrate':
            click.echo('The database needs to be migrated.')
            click.echo('Run `pixi migrate`.')
            sys.exit(1)


def calculate_migrations_needed():
    migrations = _find_migrations()
    version = _get_version()

    needed = []
    for mig in migrations:
        if mig.version > version:
            needed.append(mig)
    return sorted(needed, key=lambda m: m.version)


def _find_migrations():
    migrations = []
    for sql_path in MIGRATIONS_DIR.glob('*.sql'):
        try:
            migrations.append(
                Migration(path=sql_path, version=int(sql_path.stem))
            )
        except ValueError:
            click.echo(f'Invalid migration name: {sql_path}.')
            raise click.Abort
    return migrations


def _get_version():
    with database() as (conn, cursor):
        cursor.execute('SELECT MAX(version) FROM versions')
        return cursor.fetchone()[0] or 0



import os
from configparser import ConfigParser

from pixi import CONFIG_DIR
from pixi.errors import InvalidConfig

CONFIG_PATH = CONFIG_DIR / 'config.ini'

DEFAULT_CONFIG = {
    'pixi': {
        'refresh_token': '',
        'download_directory': '',
    },
}


def write_default_config():
    if not CONFIG_PATH.exists():
        parser = ConfigParser()
        parser.read_dict(DEFAULT_CONFIG)
        with CONFIG_PATH.open('w') as f:
            parser.write(f)


class Config:
    """
    A "proxy singleton" that returns the same ConfigParser instance when
    instantiated.
    """

    __parser = None

    def __new__(cls, validate=True):
        if cls.__parser is None:
            cls.__parser = _load_config()
            if validate:
                _validate_config(cls.__parser)
        return cls.__parser


def _load_config():
    ConfigParser.save = _save_config
    config = ConfigParser()
    config.read(CONFIG_DIR / 'config.ini')
    return config


def _save_config(parser):
    with CONFIG_PATH.open('w') as f:
        parser.write(f)


def _validate_config(config):
    if not config:
        raise InvalidConfig('Empty file')

    try:
        dl_dir = config['pixi']['download_directory']
        if not os.path.isdir(dl_dir) or not os.access(dl_dir, os.W_OK):
            raise InvalidConfig(
                'Download directory does not exist or is not writeable'
            )
    except KeyError:
        raise InvalidConfig('Download directory not configured')



import os
from urllib import parse

import click
from pixivapi import BadApiResponse, Size
from requests import RequestException

from pixi.database import database
from pixi.errors import DownloadFailed, DuplicateImage, InvalidURL, PixiError


def parse_id(string, path=None, param=None):
    try:
        return int(string)
    except ValueError:
        pass

    if path and param:
        try:
            split = parse.urlsplit(string)
            if split.netloc == 'www.pixiv.net' and split.path == path:
                return int(dict(parse.parse_qsl(split.query))[param])
        except (KeyError, TypeError):
            pass

    raise InvalidURL


def rename_duplicate_file(path):
    new_path = path
    dupe_number = 1
    while new_path.exists():
        name, ext = os.path.splitext(path.name)
        new_path = path.with_name(f'{name} ({dupe_number}){ext}')
        dupe_number += 1
    return new_path


def format_filename(id_, title):
    return f'{id_}. {title}'


def resolve_track_download(track_download, directory):
    if track_download is None:
        return not directory
    return track_download


def download_image(
    illustration,
    directory,
    tries=1,
    allow_duplicate=False,
    track_download=True,
):
    if not allow_duplicate:
        try:
            check_duplicate(illustration.id)
        except DuplicateImage:
            return click.echo(
                f'{illustration.id}. {illustration.title} has been downloaded '
                'previously, skipping...'
            )

    for attempt in range(tries):
        try:
            filename = format_filename(illustration.id, illustration.title)

            if illustration.meta_pages:
                click.echo(
                    'Downloading multi-page illustration '
                    f'{illustration.id}. {illustration.title}.'
                )
            else:
                click.echo(
                    f'Downloading illustration {illustration.id}. '
                    f'{illustration.title}.'
                )

            illustration.download(
                directory=directory,
                size=Size.ORIGINAL,
                filename=filename,
            )

            click.echo()
            clear_failed(illustration.id)
            if track_download:
                record_download(illustration.id, str(directory))

            break
        except (BadApiResponse, RequestException) as e:
            click.echo(
                f'Failed to download illustration {illustration.id}. '
                f'{illustration.title} ({e}). Attempting to re-download '
                f'(attempt {attempt + 1}).'
            )
    else:
        mark_failed(illustration)
        raise DownloadFailed


def download_pages(
    get_next_response,
    starting_offset,
    directory,
    allow_duplicates=False,
    track_download=True,
    start_page=1,
):
    response = get_next_response(starting_offset)
    if not response['illustrations']:
        raise PixiError('No illustrations found.')

    page = start_page
    while True:
        click.echo(f'Downloading page {page} of illustrations.\n')
        for illustration in response['illustrations']:
            try:
                download_image(
                    illustration,
                    directory=directory,
                    tries=3,
                    allow_duplicate=allow_duplicates,
                    track_download=(
                        resolve_track_download(track_download, directory)
                    ),
                )
            except DownloadFailed:
                click.echo(
                    f'Failed to download image {illustration.id}. '
                    f'{illustration.title} three times. Skipping...'
                )

        if not response['next']:
            break

        response = get_next_response(response['next'])
        page += 1


def mark_failed(illustration):
    with database() as (conn, cursor):
        cursor.execute(
            """
            INSERT OR IGNORE INTO failed (id, artist, title) VALUES (?, ?, ?)
            """,
            (
                illustration.id,
                illustration.user.account,
                illustration.title,
            )
        )


def clear_failed(illustration_id):
    with database() as (conn, cursor):
        cursor.execute(
            """
            DELETE FROM failed WHERE id = ?
            """,
            (
                illustration_id,
            )
        )


def record_download(illustration_id, path):
    with database() as (conn, cursor):
        cursor.execute(
            """
            INSERT INTO downloaded (id, path) VALUES (?, ?)
            """,
            (
                illustration_id,
                path,
            )
        )


def check_duplicate(illustration_id):
    with database() as (conn, cursor):
        cursor.execute(
            """
            SELECT path FROM downloaded WHERE id = ?
            """,
            (
                illustration_id,
            )
        )
        row = cursor.fetchone()
        if row:
            raise DuplicateImage(row['path'])



import traceback

import click

import pixi.commands  # noqa
from pixi import commandgroup, make_app_directories
from pixi.config import write_default_config
from pixi.database import (
    confirm_database_is_updated,
    create_database_if_nonexistent,
)
from pixi.errors import (
    DownloadFailed,
    GoAuthenticate,
    InvalidConfig,
    PixiError,
)


def run():
    try:
        make_app_directories()
        write_default_config()
        create_database_if_nonexistent()
        confirm_database_is_updated()
        commandgroup()
    except GoAuthenticate:
        click.echo('Invalid token. Re-authenticate with `pixi auth`.')
    except InvalidConfig as e:
        click.echo(f'Invalid config: {e}. Edit the config with `pixi config`.')
    except DownloadFailed:
        click.echo('Failed to download image.\n')
        traceback.print_exc()
    except PixiError as e:
        click.echo(e)


run()



import functools

import click


def download_directory(func):
    return functools.wraps(func)(
        click.option(
            '--directory', '-d',
            type=click.Path(exists=True, file_okay=False),
            help='Config override for download directory.',
        )(func)
    )


def allow_duplicates(func):
    return functools.wraps(func)(
        click.option(
            '--allow-duplicates', '-a',
            is_flag=True,
            default=False,
            help='Downloads illustrations even if previously downloaded.',
        )(func)
    )


def track_download(func):
    return functools.wraps(func)(
        click.option(
            '--track/--no-track', '-t/-T',
            default=None,
            help='Record the downloaded image to avoid future duplicates.',
        )(func)
    )


def page(func):
    return functools.wraps(func)(
        click.option(
            '--page', '-p',
            nargs=1,
            type=click.INT,
            default=1,
            help='Page to start downloading on.',
        )(func)
    )


def visibility(func):
    return functools.wraps(func)(
        click.option(
            '--visibility', '-v',
            type=click.Choice(['public', 'private']),
            help=(
                'The visibility of the bookmarks that should be downloaded. '
                'Leave blank to download all.'
            ),
        )(func)
    )



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['pixi']

package_data = \
{'': ['*'], 'pixi': ['migrations/*']}

install_requires = \
['appdirs>=1.4,<2.0',
 'click>=7.0,<8.0',
 'pixiv-api>=0.1.1,<0.2.0',
 'tqdm>=4.32,<5.0']

entry_points = \
{'console_scripts': ['pixi = pixi.__main__:run']}

setup_kwargs = {
    'name': 'pixi',
    'version': '0.1.0',
    'description': 'A command line tool to download images from Pixiv.',
    'long_description': '# pixi\n\n[![Build Status](https://travis-ci.org/dazuling/pixi.svg?branch=master)](https://travis-ci.org/dazuling/pixi)\n[![Coverage Status](https://coveralls.io/repos/github/dazuling/pixi/badge.svg?branch=master)](https://coveralls.io/github/dazuling/pixi?branch=master)\n[![Pypi](https://img.shields.io/pypi/v/pixi.svg)](https://pypi.python.org/pypi/pixi)\n[![Pyversions](https://img.shields.io/pypi/pyversions/pixi.svg)](https://pypi.python.org/pypi/pixi)\n\nA command line tool to download illustrations from Pixiv.\n\n```\nUsage: pixi [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  artist     Download illustrations of an artist by URL or ID.\n  auth       Log into Pixiv and generate a refresh token.\n  bookmarks  Download illustrations bookmarked by a user.\n  config     Edit the config file.\n  failed     View illustrations that failed to download.\n  illust     Download an illustration by URL or ID.\n  migrate    Upgrade the database to the latest migration.\n  wipe       Wipe the saved history of downloaded illustrations.\n```\n\n## Usage\n\nThis tool can be installed from PyPI as `pixi`.\n\n```sh\n$ pip install --user pixi\n```\n\nAfter installation, create the database and configure pixi with the following\ncommands.\n\n```sh\n$ pixi migrate  # Migrate the database\n$ pixi config  # Configure pixi\n```\n\nRefer to the [configuration section](#Configuration) for details on the various\nconfiguration options.\n\nNow you can begin downloading!\n\nFor example, the following commands download an illustration. pixi accepts both\na URL to the illustration as well as just the illustration ID. The same applies\nto all inputs that accept ID values.\n\n```sh\n$ pixi illustration https://www.pixiv.net/member_illust.php?mode=medium&illust_id=64930973\n```\n\n```sh\n$ pixi illustration 64930973\n```\n\nDownloading all the illustrations of an artist can be done with the following\ncommand.\n\n```sh\n$ pixi artist https://www.pixiv.net/member.php?id=2188232\n```\n\nBookmarks, public and private, can be downloaded with the following command.\n\n```sh\n$ pixi bookmarks\n```\n\nThe public bookmarks of other users can also be downloaded.\n\n```sh\n$ pixi bookmarks --user https://www.pixiv.net/member.php?id=2188232\n```\n\nAnd the following command downloads all bookmarks matching a user-assigned\nbookmark tag.\n\n```sh\n$ pixi bookmarks --tag "has cats"\n```\n\nTo view all the options available to a specific command, run the command with\nthe `--help` flag. For example, `illustration`\'s options can be viewed with the\nfollowing command.\n\n```sh\n$ pixi --help illustration\n```\n\nWhen downloading many images from an artist or a user\'s bookmarks, an image\ncan occasionally fail to download. If an image fails to download after several\nretries, it will be recorded and skipped. Failed images can be viewed with the\nfollowing command.\n\n```sh\n$ pixi failed\n```\n\nIf an image on the failed list is successfully downloaded, it will\nautomatically be removed from the list. To wipe the entire failed list, the\nfollowing command should be run.\n\n```sh\n$ pixi wipe --table=failed\n```\n\npixi also keeps track of which illustrations have been downloaded and will avoid\ndownloading duplicate illustrations. However, if you wish to re-download\nillustrations, pass the `--allow-duplicates` (or `-a`) flag.\n\nBy default, illustration downloads will be tracked if they are downloaded to\nthe default downloads directory and not tracked if they aren\'t. This behavior\ncan be manually set with the `--track/--no-track` (or `-t/-T`) flag.\n\nIf you wish to wipe the database of tracked downloads, run the following\ncommand and confirm the action.\n\n```sh\n$ pixi wipe --table=downloads\n```\n\n## Configuration\n\nThe configuration file is in `ini` format. A demo configuration is included\nbelow. To run pixi, a default download directory must be configured.\n\n```ini\n[pixi]\n; Leave this blank; the script will auto-populate it.\nrefresh_token =\n; The default directory for iillustrations to be downloaded to.\ndownload_directory = /home/dazzler/images/pixiv\n```\n',
    'author': 'dazzler',
    'author_email': 'dazzler@riseup.net',
    'url': 'https://github.com/dazuling/pixi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import sys
from datetime import datetime
from pathlib import Path

import click
from pixivapi import BadApiResponse, LoginError, Visibility
from requests import RequestException

from pixi import commandgroup
from pixi.client import Client
from pixi.config import CONFIG_PATH, Config
from pixi.database import calculate_migrations_needed, database
from pixi.errors import DownloadFailed, PixiError
from pixi.options import (
    allow_duplicates,
    download_directory,
    page,
    track_download,
    visibility,
)
from pixi.util import (
    download_image,
    download_pages,
    parse_id,
    resolve_track_download,
)


@commandgroup.command()
@click.option('--username', '-u', prompt='Username')
@click.option('--password', '-p', prompt='Password', hide_input=True)
def auth(username, password):
    """Log into Pixiv and generate a refresh token."""
    client = Client(authenticate=False)

    try:
        client.login(username, password)
    except LoginError:
        raise PixiError('Invalid credentials.')

    config = Config(validate=False)
    config['pixi']['refresh_token'] = client.refresh_token
    config.save()

    click.echo('Successfully authenticated; token written to config.')


@commandgroup.command()
def config():
    """Edit the config file."""
    with CONFIG_PATH.open('r+') as f:
        text = click.edit(f.read())
        if text:
            f.seek(0)
            f.truncate(0)
            f.write(text)
            click.echo('Edit completed.')
        else:
            click.echo('Edit aborted.')


@commandgroup.command()
def migrate():
    """Upgrade the database to the latest migration."""
    migrations_needed = calculate_migrations_needed()

    if not migrations_needed:
        click.echo('Database is up to date.')
        sys.exit(1)

    with database() as (conn, cursor):
        for mig in migrations_needed:
            with mig.path.open() as sql:
                cursor.executescript(sql.read())
                cursor.execute(
                    'INSERT INTO versions (version) VALUES (?)',
                    (mig.version, ),
                )
            conn.commit()


@commandgroup.command()
@click.argument(
    'illustration',
    nargs=1,
    callback=lambda ctx, param, value: (
        parse_id(value, path='/member_illust.php', param='illust_id')
    ),
)
@download_directory
@allow_duplicates
@track_download
def illust(illustration, directory, allow_duplicates, track):
    """Download an illustration by URL or ID."""
    try:
        download_image(
            Client().fetch_illustration(illustration),
            directory=(
                Path(directory or Config()['pixi']['download_directory'])
            ),
            allow_duplicate=allow_duplicates,
            track_download=resolve_track_download(track, directory),
        )
    except (BadApiResponse, RequestException) as e:
        raise DownloadFailed from e


@commandgroup.command()
@click.argument(
    'artist',
    nargs=1,
    callback=lambda ctx, param, value: (
        parse_id(value, path='/member.php', param='id')
    ),
)
@page
@download_directory
@allow_duplicates
@track_download
def artist(artist, page, directory, allow_duplicates, track):
    """Download illustrations of an artist by URL or ID."""
    client = Client()

    def get_next_response(offset):
        return client.fetch_user_illustrations(artist, offset=offset)

    download_pages(
        get_next_response,
        starting_offset=(page - 1) * 30,
        directory=(
            Path(directory or Config()['pixi']['download_directory'])
        ),
        allow_duplicates=allow_duplicates,
        track_download=track,
    )

    click.echo(f'Finished downloading artist {artist}.')


@commandgroup.command()
@click.option(
    '--user', '-u',
    help='The user whose bookmarks to download. Default logged in account.',
    callback=lambda ctx, param, value: (
        parse_id(value, path='/member.php', param='id') if value else None
    ),
)
@click.option(
    '--tag', '-g',
    help='The bookmark tag to filter bookmarks by.',
)
@visibility
@page
@download_directory
@allow_duplicates
@track_download
def bookmarks(
    user, tag, visibility, page, directory, allow_duplicates, track
):
    """Download illustrations bookmarked by a user."""
    client = Client()

    if visibility:
        visibilities = [Visibility(visibility)]
    else:
        visibilities = [Visibility.PUBLIC, Visibility.PRIVATE]

    for visi in visibilities:
        click.echo(f'Downloading {visi.value} bookmarks.\n')

        def get_next_response(offset):
            return client.fetch_user_bookmarks(
                user_id=user or client.account.id,
                max_bookmark_id=offset,
                visibility=visi,
                tag=tag,
            )

        download_pages(
            get_next_response,
            starting_offset=(page - 1) * 30,
            directory=(
                Path(directory or Config()['pixi']['download_directory'])
            ),
            allow_duplicates=allow_duplicates,
            track_download=track,
        )

    click.echo(f'Finished downloading artist {artist}.')


@commandgroup.command()
def failed():
    """View illustrations that failed to download."""
    with database() as (conn, cursor):
        cursor.execute(
            """
            SELECT
                id,
                artist,
                title,
                time
            FROM failed
            ORDER BY time DESC
            """
        )
        for row in cursor.fetchall():
            time = datetime.fromisoformat(row['time']).strftime(
                '%b %d, %Y %H:%M:%S'
            )
            click.echo(f'{time} | {row["artist"]} - {row["title"]}')
            click.echo(
                'URL: https://www.pixiv.net/member_illust.php?mode=medium'
                f'&illust_id={row["id"]}\n'
            )


@commandgroup.command()
@click.option(
    '--table', '-t',
    type=click.Choice(['downloaded', 'failed', 'all']),
    required=True,
    help='The table to wipe.',
)
def wipe(table):
    """Wipe the saved history of downloaded illustrations."""
    tables = ['downloaded', 'failed'] if table == 'all' else [table]
    with database() as (conn, cursor):
        for t in tables:
            _confirm_table_wipe(t)
            cursor.execute(f'DELETE FROM {t}')
            click.echo(f'Wiped the {t} table.')
        conn.commit()


def _confirm_table_wipe(table):
    confirmation = click.prompt(
        f'Enter "{table}" to confirm that you wish to wipe {table}'
    )
    if confirmation != table:
        raise click.Abort
